@@ -3,8 +3,8 @@
  * 
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-9 Bruce Allen <smartmontools-support@lists.sourceforge.net>
- * Copyright (C) 2008-9 Christian Franke <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-10 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2008-10 Christian Franke <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  * Copyright (C) 2000 Andre Hedrick <andre@linux-ide.org>
  *
@@ -37,7 +37,7 @@
 #include "utility.h"
 #include "dev_ata_cmd_set.h" // for parsed_ata_device
 
-const char * atacmds_cpp_cvsid = "$Id: atacmds.cpp 3046 2010-01-22 21:30:02Z chrfranke $"
+const char * atacmds_cpp_cvsid = "$Id: atacmds.cpp 3185 2010-10-15 22:23:53Z chrfranke $"
                                  ATACMDS_H_CVSID;
 
 // for passing global control variables
@@ -58,7 +58,8 @@ extern smartmonctrl *con;
 // that SMART was first added into the ATA/ATAPI-3 Standard with
 // Revision 3 of the document, July 25, 1995.  Look at the "Document
 // Status" revision commands at the beginning of
-// http://www.t13.org/project/d2008r6.pdf to see this.
+// http://www.t13.org/Documents/UploadedDocuments/project/d2008r7b-ATA-3.pdf
+// to see this.
 #define NOVAL_0                 0x0000
 #define NOVAL_1                 0xffff
 /* word 81: minor version number */
@@ -152,7 +153,13 @@ unsigned char get_unc_attr_id(bool offline, const ata_vendor_attr_defs & defs,
                               bool & increase)
 {
   unsigned char id = (!offline ? 197 : 198);
-  increase = !!(defs[id].flags & ATTRFLAG_INCREASING);
+  const ata_vendor_attr_defs::entry & def = defs[id];
+  if (def.flags & ATTRFLAG_INCREASING)
+    increase = true; // '-v 19[78],increasing' option
+  else if (def.name.empty() || (id == 198 && def.name == "Offline_Scan_UNC_SectCt"))
+    increase = false; // no or '-v 198,offlinescanuncsectorct' option
+  else
+    id = 0; // other '-v 19[78],...' option
   return id;
 }
 
@@ -201,10 +208,12 @@ const format_name_entry format_names[] = {
   {"hex64"          , RAWFMT_HEX64},
   {"raw16(raw16)"   , RAWFMT_RAW16_OPT_RAW16},
   {"raw16(avg16)"   , RAWFMT_RAW16_OPT_AVG16},
-  {"raw24/raw24"    , RAWFMT_RAW24_RAW24},
+  {"raw24/raw24"    , RAWFMT_RAW24_DIV_RAW24},
+  {"raw24/raw32"    , RAWFMT_RAW24_DIV_RAW32},
   {"sec2hour"       , RAWFMT_SEC2HOUR},
   {"min2hour"       , RAWFMT_MIN2HOUR},
   {"halfmin2hour"   , RAWFMT_HALFMIN2HOUR},
+  {"msec24hour32"   , RAWFMT_MSEC24_HOUR32},
   {"tempminmax"     , RAWFMT_TEMPMINMAX},
   {"temp10x"        , RAWFMT_TEMP10X},
 };
@@ -222,7 +231,7 @@ const char * map_old_vendor_opts[][2] = {
   {"194,10xCelsius"               , "194,temp10x,Temperature_Celsius_x10"},
   {"194,unknown"                  , "194,raw48,Unknown_Attribute"},
   {"197,increasing"               , "197,raw48+,Total_Pending_Sectors"}, // '+' sets flag
-  {"198,offlinescanuncsectorct"   , "198,raw48,Offline_Scan_UNC_SectCt"},
+  {"198,offlinescanuncsectorct"   , "198,raw48,Offline_Scan_UNC_SectCt"}, // see also get_unc_attr_id() above
   {"198,increasing"               , "198,raw48+,Total_Offl_Uncorrectabl"}, // '+' sets flag
   {"200,writeerrorcount"          , "200,raw48,Write_Error_Count"},
   {"201,detectedtacount"          , "201,raw48,Detected_TA_Count"},
@@ -271,6 +280,19 @@ bool parse_attribute_def(const char * opt, ata_vendor_attr_defs & defs,
     flags = ATTRFLAG_INCREASING;
   }
 
+  // Split "format[:byteorder]"
+  char byteorder[8+1] = "";
+  if (strchr(fmtname, ':')) {
+    if (!(   sscanf(fmtname, "%*[^:]%n:%8[012345rvwz]%n", &n1, byteorder, &n2) >= 1
+          && n2 == (int)strlen(fmtname)))
+      return false;
+    fmtname[n1] = 0;
+    if (strchr(byteorder, 'v'))
+      flags |= (ATTRFLAG_NO_NORMVAL|ATTRFLAG_NO_WORSTVAL);
+    if (strchr(byteorder, 'w'))
+      flags |= ATTRFLAG_NO_WORSTVAL;
+  }
+
   // Find format name
   for (i = 0; ; i++) {
     if (i >= num_format_names)
@@ -280,9 +302,9 @@ bool parse_attribute_def(const char * opt, ata_vendor_attr_defs & defs,
   }
   ata_attr_raw_format format = format_names[i].format;
 
-  // 64-bit formats use the normalized value bytes.
-  if (format == RAWFMT_RAW64 || format == RAWFMT_HEX64)
-    flags |= ATTRFLAG_NO_NORMVAL;
+  // 64-bit formats use the normalized and worst value bytes.
+  if (!*byteorder && (format == RAWFMT_RAW64 || format == RAWFMT_HEX64))
+    flags |= (ATTRFLAG_NO_NORMVAL|ATTRFLAG_NO_WORSTVAL);
 
   if (!id) {
     // "N,format" -> set format for all entries
@@ -294,6 +316,7 @@ bool parse_attribute_def(const char * opt, ata_vendor_attr_defs & defs,
       defs[i].priority = priority;
       defs[i].raw_format = format;
       defs[i].flags = flags;
+      strcpy(defs[i].byteorder, byteorder);
     }
   }
   else if (defs[id].priority <= priority) {
@@ -303,6 +326,7 @@ bool parse_attribute_def(const char * opt, ata_vendor_attr_defs & defs,
     defs[id].raw_format = format;
     defs[id].priority = priority;
     defs[id].flags = flags;
+    strcpy(defs[id].byteorder, byteorder);
   }
 
   return true;
@@ -317,7 +341,7 @@ std::string create_vendor_attribute_arg_list()
   std::string s;
   unsigned i;
   for (i = 0; i < num_format_names; i++)
-    s += strprintf("%s\tN,%s[,ATTR_NAME]",
+    s += strprintf("%s\tN,%s[:012345rvwz][,ATTR_NAME]",
       (i>0 ? "\n" : ""), format_names[i].name);
   for (i = 0; i < num_old_vendor_opts; i++)
     s += strprintf("\n\t%s", map_old_vendor_opts[i][0]);
@@ -420,12 +444,19 @@ void print_regs(const char * prefix, const ata_out_regs & r, const char * suffix
 static void prettyprint(const unsigned char *p, const char *name){
   pout("\n===== [%s] DATA START (BASE-16) =====\n", name);
   for (int i=0; i<512; i+=16, p+=16)
+#define P(n) (isprint((int)(p[n]))?(int)(p[n]):'.')
     // print complete line to avoid slow tty output and extra lines in syslog.
     pout("%03d-%03d: %02x %02x %02x %02x %02x %02x %02x %02x "
-                    "%02x %02x %02x %02x %02x %02x %02x %02x\n",
+                    "%02x %02x %02x %02x %02x %02x %02x %02x"
+                    " |%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c|"
+         "%c",
          i, i+16-1,
          p[ 0], p[ 1], p[ 2], p[ 3], p[ 4], p[ 5], p[ 6], p[ 7],
-         p[ 8], p[ 9], p[10], p[11], p[12], p[13], p[14], p[15]);
+         p[ 8], p[ 9], p[10], p[11], p[12], p[13], p[14], p[15], 
+         P( 0), P( 1), P( 2), P( 3), P( 4), P( 5), P( 6), P( 7),
+         P( 8), P( 9), P(10), P(11), P(12), P(13), P(14), P(15),
+         '\n');
+#undef P
   pout("===== [%s] DATA END (512 Bytes) =====\n\n", name);
 }
 
@@ -600,8 +631,10 @@ int smartcommandhandler(ata_device * device, smart_command_set command, int sele
                  "probable SAT/USB truncation\n");
         } else {
           // We haven't gotten output that makes sense; print out some debugging info
-          pout("Error SMART Status command failed\n"
-               "Please get assistance from %s\n", PACKAGE_HOMEPAGE);
+          pout("Error SMART Status command failed\n");
+          pout("Please get assistance from %s\n", PACKAGE_HOMEPAGE);
+          pout("Register values returned from SMART Status command are:\n");
+          print_regs(" ", out.out_regs);
           errno = EIO;
           retval = -1;
         }
@@ -837,7 +870,7 @@ int ataVersionInfo(const char ** description, const ata_identify_device * drive,
   // First check if device has ANY ATA version information in it
   if (major==NOVAL_0 || major==NOVAL_1) {
     *description=NULL;
-    return -1;
+    return 0; // No info found
   }
   
   // The minor revision number has more information - try there first
@@ -1526,7 +1559,7 @@ int ataSmartTest(ata_device * device, int testtype, const ata_selective_selftest
   else if ((select=(testtype==SELECTIVE_SELF_TEST || testtype==SELECTIVE_CAPTIVE_SELF_TEST)))
     type="Selective self-test";
   else
-    type="[Unrecognized] self-test";
+    type = 0;
   
   // If doing a selective self-test, first use WRITE_LOG to write the
   // selective self-test log.
@@ -1540,6 +1573,8 @@ int ataSmartTest(ata_device * device, int testtype, const ata_selective_selftest
   //  Print ouf message that we are sending the command to test
   if (testtype==ABORT_SELF_TEST)
     sprintf(cmdmsg,"Abort SMART off-line mode self-test routine");
+  else if (!type)
+    sprintf(cmdmsg, "SMART EXECUTE OFF-LINE IMMEDIATE subcommand 0x%02x", testtype);
   else
     sprintf(cmdmsg,"Execute SMART %s routine immediately in %s mode",type,captive);
   pout("Sending command: \"%s\".\n",cmdmsg);
@@ -1567,8 +1602,11 @@ int ataSmartTest(ata_device * device, int testtype, const ata_selective_selftest
   // Since the command succeeded, tell user
   if (testtype==ABORT_SELF_TEST)
     pout("Self-testing aborted!\n");
-  else
-    pout("Drive command \"%s\" successful.\nTesting has begun.\n",cmdmsg);
+  else {
+    pout("Drive command \"%s\" successful.\n", cmdmsg);
+    if (type)
+      pout("Testing has begun.\n");
+  }
   return 0;
 }
 
@@ -1705,8 +1743,10 @@ int isSupportSelectiveSelfTest(const ata_smart_values * data)
 
 // Get attribute state
 ata_attr_state ata_get_attr_state(const ata_smart_attribute & attr,
-                                  const ata_smart_threshold_entry & thre,
-                                  const ata_vendor_attr_defs & defs)
+                                  int attridx,
+                                  const ata_smart_threshold_entry * thresholds,
+                                  const ata_vendor_attr_defs & defs,
+                                  unsigned char * threshval /* = 0 */)
 {
   if (!attr.id)
     return ATTRSTATE_NON_EXISTING;
@@ -1717,27 +1757,35 @@ ata_attr_state ata_get_attr_state(const ata_smart_attribute & attr,
   if (defs[attr.id].flags & ATTRFLAG_NO_NORMVAL)
     return ATTRSTATE_NO_NORMVAL;
 
-  // No threshold if thresholds cannot be read.
-  if (!thre.id && !thre.threshold)
-    return ATTRSTATE_NO_THRESHOLD;
+  // Normally threshold is at same index as attribute
+  int i = attridx;
+  if (thresholds[i].id != attr.id) {
+    // Find threshold id in table
+    for (i = 0; thresholds[i].id != attr.id; ) {
+      if (++i >= NUMBER_ATA_SMART_ATTRIBUTES)
+        // Threshold id missing or thresholds cannot be read
+        return ATTRSTATE_NO_THRESHOLD;
+    }
+  }
+  unsigned char threshold = thresholds[i].threshold;
 
-  // Bad threshold if id's don't match
-  if (attr.id != thre.id)
-    return ATTRSTATE_BAD_THRESHOLD;
+  // Return threshold if requested
+  if (threshval)
+    *threshval = threshold;
 
   // Don't report a failed attribute if its threshold is 0.
   // ATA-3 (X3T13/2008D Revision 7b) declares 0x00 as the "always passing"
   // threshold (Later ATA versions declare all thresholds as "obsolete").
   // In practice, threshold value 0 is often used for usage attributes.
-  if (!thre.threshold)
+  if (!threshold)
     return ATTRSTATE_OK;
 
   // Failed now if current value is below threshold
-  if (attr.current <= thre.threshold)
+  if (attr.current <= threshold)
     return ATTRSTATE_FAILED_NOW;
 
-  // Failed in the passed if worst value is below threshold
-  if (attr.worst <= thre.threshold)
+  // Failed in the past if worst value is below threshold
+  if (!(defs[attr.id].flags & ATTRFLAG_NO_WORSTVAL) && attr.worst <= threshold)
     return ATTRSTATE_FAILED_PAST;
 
   return ATTRSTATE_OK;
@@ -1767,24 +1815,42 @@ static ata_attr_raw_format get_default_raw_format(unsigned char id)
 uint64_t ata_get_attr_raw_value(const ata_smart_attribute & attr,
                                 const ata_vendor_attr_defs & defs)
 {
-  // Get 48 bit raw value
-  const unsigned char * raw = attr.raw;
-  uint64_t rawvalue;
-  rawvalue =              raw[0]
-             | (          raw[1] <<  8)
-             | (          raw[2] << 16)
-             | ((uint64_t)raw[3] << 24)
-             | ((uint64_t)raw[4] << 32)
-             | ((uint64_t)raw[5] << 40);
+  const ata_vendor_attr_defs::entry & def = defs[attr.id];
 
-  if (defs[attr.id].flags & ATTRFLAG_NO_NORMVAL) {
-    // Some SSD vendors use bytes 3-10 from the Attribute
-    // Data Structure to store a 64-bit raw value.
-    rawvalue <<= 8;
-    rawvalue |= attr.worst;
-    rawvalue <<= 8;
-    rawvalue |= attr.current;
+  // Use default byteorder if not specified
+  const char * byteorder = def.byteorder;
+  if (!*byteorder) {
+    switch (def.raw_format) {
+      case RAWFMT_RAW64:
+      case RAWFMT_HEX64:
+        byteorder = "543210wv"; break;
+      case RAWFMT_RAW24_DIV_RAW32:
+      case RAWFMT_MSEC24_HOUR32:
+        byteorder = "r543210"; break;
+      default:
+        byteorder = "543210"; break;
+    }
   }
+
+  // Build 64-bit value from selected bytes
+  uint64_t rawvalue = 0;
+  for (int i = 0; byteorder[i]; i++) {
+    unsigned char b;
+    switch (byteorder[i]) {
+      case '0': b = attr.raw[0];  break;
+      case '1': b = attr.raw[1];  break;
+      case '2': b = attr.raw[2];  break;
+      case '3': b = attr.raw[3];  break;
+      case '4': b = attr.raw[4];  break;
+      case '5': b = attr.raw[5];  break;
+      case 'r': b = attr.reserv;  break;
+      case 'v': b = attr.current; break;
+      case 'w': b = attr.worst;   break;
+      default : b = 0;            break;
+    }
+    rawvalue <<= 8; rawvalue |= b;
+  }
+
   return rawvalue;
 }
 
@@ -1793,10 +1859,11 @@ uint64_t ata_get_attr_raw_value(const ata_smart_attribute & attr,
 std::string ata_format_attr_raw_value(const ata_smart_attribute & attr,
                                       const ata_vendor_attr_defs & defs)
 {
-  // Get 48 bit or64 bit raw value
+  // Get 48 bit or 64 bit raw value
   uint64_t rawvalue = ata_get_attr_raw_value(attr, defs);
 
   // Get 16 bit words
+  // TODO: rebuild raw[6] from rawvalue
   const unsigned char * raw = attr.raw;
   unsigned word[3];
   word[0] = raw[0] | (raw[1] << 8);
@@ -1821,6 +1888,7 @@ std::string ata_format_attr_raw_value(const ata_smart_attribute & attr,
     break;
 
   case RAWFMT_RAW48:
+  case RAWFMT_RAW64:
     s = strprintf("%"PRIu64, rawvalue);
     break;
 
@@ -1828,10 +1896,6 @@ std::string ata_format_attr_raw_value(const ata_smart_attribute & attr,
     s = strprintf("0x%012"PRIx64, rawvalue);
     break;
 
-  case RAWFMT_RAW64:
-    s = strprintf("%"PRIu64, rawvalue);
-    break;
-
   case RAWFMT_HEX64:
     s = strprintf("0x%016"PRIx64, rawvalue);
     break;
@@ -1848,10 +1912,14 @@ std::string ata_format_attr_raw_value(const ata_smart_attribute & attr,
       s += strprintf(" (Average %u)", word[1]);
     break;
 
-  case RAWFMT_RAW24_RAW24:
-    s = strprintf("%d/%d",
-      raw[0] | (raw[1]<<8) | (raw[2]<<16),
-      raw[3] | (raw[4]<<8) | (raw[5]<<16));
+  case RAWFMT_RAW24_DIV_RAW24:
+    s = strprintf("%u/%u",
+      (unsigned)(rawvalue >> 24), (unsigned)(rawvalue & 0x00ffffffULL));
+    break;
+
+  case RAWFMT_RAW24_DIV_RAW32:
+    s = strprintf("%u/%u",
+      (unsigned)(rawvalue >> 32), (unsigned)(rawvalue & 0xffffffffULL));
     break;
 
   case RAWFMT_MIN2HOUR:
@@ -1885,6 +1953,17 @@ std::string ata_format_attr_raw_value(const ata_smart_attribute & attr,
     }
     break;
 
+  case RAWFMT_MSEC24_HOUR32:
+    {
+      // hours + milliseconds
+      unsigned hours = (unsigned)(rawvalue & 0xffffffffULL);
+      unsigned milliseconds = (unsigned)(rawvalue >> 32);
+      unsigned seconds = milliseconds / 1000;
+      s = strprintf("%uh+%02um+%02u.%03us",
+        hours, seconds / 60, seconds % 60, milliseconds % 1000);
+    }
+    break;
+
   case RAWFMT_TEMPMINMAX:
     // Temperature
     s = strprintf("%u", word[0]);
@@ -1902,7 +1981,7 @@ std::string ata_format_attr_raw_value(const ata_smart_attribute & attr,
         unsigned t = lo; lo = hi; hi = t;
       }
       if (lo <= word[0] && word[0] <= hi)
-        s += strprintf(" (Lifetime Min/Max %u/%u)", lo, hi);
+        s += strprintf(" (Min/Max %u/%u)", lo, hi);
       else
         s += strprintf(" (%d %d %d %d)", raw[5], raw[4], raw[3], raw[2]);
     }
@@ -2180,6 +2259,13 @@ int ataReadSCTTempHist(ata_device * device, ata_sct_temperature_history_table *
   cmd.function_code = 1; // Read table
   cmd.table_id      = 2; // Temperature History Table
 
+  // swap endian order if needed
+  if (isbigendian()) {
+    swapx(&cmd.action_code);
+    swapx(&cmd.function_code);
+    swapx(&cmd.table_id);
+  }
+
   // write command via SMART log page 0xe0
   if (smartcommandhandler(device, WRITE_LOG, 0xe0, (char *)&cmd)){
     syserror("Error Write SCT Data Table command failed");
@@ -2198,7 +2284,7 @@ int ataReadSCTTempHist(ata_device * device, ata_sct_temperature_history_table *
     return -1;
 
   if (!(sts->ext_status_code == 0 && sts->action_code == 5 && sts->function_code == 1)) {
-    pout("Error unexcepted SCT status 0x%04x (action_code=%u, function_code=%u)\n",
+    pout("Error unexpected SCT status 0x%04x (action_code=%u, function_code=%u)\n",
       sts->ext_status_code, sts->action_code, sts->function_code);
     return -1;
   }
@@ -2242,6 +2328,15 @@ int ataSetSCTTempInterval(ata_device * device, unsigned interval, bool persisten
   cmd.state         = interval;
   cmd.option_flags  = (persistent ? 0x01 : 0x00);
 
+  // swap endian order if needed
+  if (isbigendian()) {
+    swapx(&cmd.action_code);
+    swapx(&cmd.function_code);
+    swapx(&cmd.feature_code);
+    swapx(&cmd.state);
+    swapx(&cmd.option_flags);
+  }
+
   // write command via SMART log page 0xe0
   if (smartcommandhandler(device, WRITE_LOG, 0xe0, (char *)&cmd)){
     syserror("Error Write SCT Feature Control Command failed");
@@ -2260,6 +2355,96 @@ int ataSetSCTTempInterval(ata_device * device, unsigned interval, bool persisten
   return 0;
 }
 
+// Get/Set SCT Error Recovery Control
+static int ataGetSetSCTErrorRecoveryControltime(ata_device * device, unsigned type,
+                                                bool set, unsigned short & time_limit)
+{
+  // Check initial status
+  ata_sct_status_response sts;
+  if (ataReadSCTStatus(device, &sts))
+    return -1;
+
+  // Do nothing if other SCT command is executing
+  if (sts.ext_status_code == 0xffff) {
+    pout("Another SCT command is executing, abort Error Recovery Control\n"
+         "(SCT ext_status_code 0x%04x, action_code=%u, function_code=%u)\n",
+      sts.ext_status_code, sts.action_code, sts.function_code);
+    return -1;
+  }
+
+  ata_sct_error_recovery_control_command cmd; memset(&cmd, 0, sizeof(cmd));
+  // CAUTION: DO NOT CHANGE THIS VALUE (SOME ACTION CODES MAY ERASE DISK)
+  cmd.action_code    = 3; // Error Recovery Control command
+  cmd.function_code  = (set ? 1 : 2); // 1=Set timer, 2=Get timer
+  cmd.selection_code = type; // 1=Read timer, 2=Write timer
+  if (set)
+    cmd.time_limit   = time_limit;
+
+  // swap endian order if needed
+  if (isbigendian()) {
+    swapx(&cmd.action_code);
+    swapx(&cmd.function_code);
+    swapx(&cmd.selection_code);
+    swapx(&cmd.time_limit);
+  }
+
+  // write command via SMART log page 0xe0
+  // TODO: Debug output
+  ata_cmd_in in;
+  in.in_regs.command = ATA_SMART_CMD;
+  in.in_regs.lba_high = SMART_CYL_HI; in.in_regs.lba_mid = SMART_CYL_LOW;
+  in.in_regs.features = ATA_SMART_WRITE_LOG_SECTOR;
+  in.in_regs.lba_low = 0xe0;
+  in.set_data_out(&cmd, 1);
+
+  if (!set)
+    // Time limit returned in ATA registers
+    in.out_needed.sector_count = in.out_needed.lba_low = true;
+
+  ata_cmd_out out;
+  if (!device->ata_pass_through(in, out)) {
+    pout("Error Write SCT Error Recovery Control Command failed: %s\n", device->get_errmsg());
+    return -1;
+  }
+
+  // re-read and check SCT status
+  if (ataReadSCTStatus(device, &sts))
+    return -1;
+
+  if (!(sts.ext_status_code == 0 && sts.action_code == 3 && sts.function_code == (set ? 1 : 2))) {
+    pout("Error unexcepted SCT status 0x%04x (action_code=%u, function_code=%u)\n",
+      sts.ext_status_code, sts.action_code, sts.function_code);
+    return -1;
+  }
+
+  if (!set) {
+    // Check whether registers are properly returned by ioctl()
+    if (!(out.out_regs.sector_count.is_set() && out.out_regs.lba_low.is_set())) {
+      // TODO: Output register support should be checked within each ata_pass_through()
+      // implementation before command is issued.
+      pout("Error SMART WRITE LOG does not return COUNT and LBA_LOW register\n");
+      return -1;
+    }
+    // Return value to caller
+    time_limit = out.out_regs.sector_count | (out.out_regs.lba_low << 8);
+  }
+
+  return 0;
+}
+
+// Get SCT Error Recovery Control
+int ataGetSCTErrorRecoveryControltime(ata_device * device, unsigned type, unsigned short & time_limit)
+{
+  return ataGetSetSCTErrorRecoveryControltime(device, type, false/*get*/, time_limit);
+}
+
+// Set SCT Error Recovery Control
+int ataSetSCTErrorRecoveryControltime(ata_device * device, unsigned type, unsigned short time_limit)
+{
+  return ataGetSetSCTErrorRecoveryControltime(device, type, true/*set*/, time_limit);
+}
+
+
 // Print one self-test log entry.
 // Returns true if self-test showed an error.
 bool ataPrintSmartSelfTestEntry(unsigned testnum, unsigned char test_type,
@@ -2268,7 +2453,7 @@ bool ataPrintSmartSelfTestEntry(unsigned testnum, unsigned char test_type,
                                 uint64_t failing_lba,
                                 bool print_error_only, bool & print_header)
 {
-  const char * msgtest;
+  std::string msgtest;
   switch (test_type) {
     case 0x00: msgtest = "Offline";            break;
     case 0x01: msgtest = "Short offline";      break;
@@ -2282,13 +2467,13 @@ bool ataPrintSmartSelfTestEntry(unsigned testnum, unsigned char test_type,
     case 0x84: msgtest = "Selective captive";  break;
     default:
       if ((0x40 <= test_type && test_type <= 0x7e) || 0x90 <= test_type)
-        msgtest = "Vendor offline";
+        msgtest = strprintf("Vendor (0x%02x)", test_type);
       else
-        msgtest = "Reserved offline";
+        msgtest = strprintf("Reserved (0x%02x)", test_type);
   }
 
   bool is_error = false;
-  const char * msgstat;
+  std::string msgstat;
   switch (test_status >> 4) {
     case 0x0: msgstat = "Completed without error";       break;
     case 0x1: msgstat = "Aborted by host";               break;
@@ -2300,7 +2485,7 @@ bool ataPrintSmartSelfTestEntry(unsigned testnum, unsigned char test_type,
     case 0x7: msgstat = "Completed: read failure";       is_error = true; break;
     case 0x8: msgstat = "Completed: handling damage??";  is_error = true; break;
     case 0xf: msgstat = "Self-test routine in progress"; break;
-    default:  msgstat = "Unknown/reserved test status";
+    default:  msgstat = strprintf("Unknown status (0x%x)", test_status >> 4);
   }
 
   if (!is_error && print_error_only)
@@ -2318,8 +2503,8 @@ bool ataPrintSmartSelfTestEntry(unsigned testnum, unsigned char test_type,
   else
     strcpy(msglba, "-");
 
-  pout("#%2u  %-19s %-29s %1d0%%  %8u         %s\n", testnum, msgtest, msgstat,
-       test_status & 0x0f, timestamp, msglba);
+  pout("#%2u  %-19s %-29s %1d0%%  %8u         %s\n", testnum,
+       msgtest.c_str(), msgstat.c_str(), test_status & 0x0f, timestamp, msglba);
 
   return is_error;
 }
