@@ -26,7 +26,7 @@
 #ifndef ATACMDS_H_
 #define ATACMDS_H_
 
-#define ATACMDS_H_CVSID "$Id: atacmds.h 2983 2009-11-14 21:41:41Z chrfranke $"
+#define ATACMDS_H_CVSID "$Id: atacmds.h 3133 2010-08-18 17:23:27Z chrfranke $"
 
 #include "dev_interface.h" // ata_device
 
@@ -543,6 +543,20 @@ struct ata_sct_status_response
 #pragma pack()
 ASSERT_SIZEOF_STRUCT(ata_sct_status_response, 512);
 
+// SCT Error Recovery Control command (send with SMART_WRITE_LOG page 0xe0)
+// Table 88 of T13/1699-D Revision 6a
+#pragma pack(1)
+struct ata_sct_error_recovery_control_command
+{
+  unsigned short action_code;       // 3 = Error Recovery Control
+  unsigned short function_code;     // 1 = Set, 2 = Return
+  unsigned short selection_code;    // 1 = Read Timer, 2 = Write Timer
+  unsigned short time_limit;        // If set: Recovery time limit in 100ms units
+  unsigned short words004_255[252]; // reserved
+} ATTR_PACKED;
+#pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_sct_error_recovery_control_command, 512);
+
 // SCT Feature Control command (send with SMART_WRITE_LOG page 0xe0)
 // Table 72 of T13/1699-D Revision 3f
 #pragma pack(1)
@@ -552,8 +566,8 @@ struct ata_sct_feature_control_command
   unsigned short function_code;     // 1 = Set, 2 = Return, 3 = Return options
   unsigned short feature_code;      // 3 = Temperature logging interval
   unsigned short state;             // Interval
-  unsigned short option_flags;      // Bit 0: persistent, Bits 1-31: reserved
-  unsigned short words005_255[251]; // reserved 
+  unsigned short option_flags;      // Bit 0: persistent, Bits 1-15: reserved
+  unsigned short words005_255[251]; // reserved
 } ATTR_PACKED;
 #pragma pack()
 ASSERT_SIZEOF_STRUCT(ata_sct_feature_control_command, 512);
@@ -646,18 +660,21 @@ enum ata_attr_raw_format
   RAWFMT_HEX64,
   RAWFMT_RAW16_OPT_RAW16,
   RAWFMT_RAW16_OPT_AVG16,
-  RAWFMT_RAW24_RAW24,
+  RAWFMT_RAW24_DIV_RAW24,
+  RAWFMT_RAW24_DIV_RAW32,
   RAWFMT_SEC2HOUR,
   RAWFMT_MIN2HOUR,
   RAWFMT_HALFMIN2HOUR,
+  RAWFMT_MSEC24_HOUR32,
   RAWFMT_TEMPMINMAX,
   RAWFMT_TEMP10X,
 };
 
 // Attribute flags
 enum {
-  ATTRFLAG_INCREASING = 0x01, // Value not reset (for reallocated/pending counts)
-  ATTRFLAG_NO_NORMVAL = 0x02  // Normalized value not valid
+  ATTRFLAG_INCREASING = 0x01,   // Value not reset (for reallocated/pending counts)
+  ATTRFLAG_NO_NORMVAL = 0x02,   // Normalized value not valid
+  ATTRFLAG_NO_WORSTVAL = 0x04   // Worst value not valid
 };
 
 // Vendor attribute display defs for all attribute ids
@@ -670,12 +687,13 @@ public:
     ata_attr_raw_format raw_format; // Raw value print format
     ata_vendor_def_prior priority; // Setting priority
     unsigned flags; // ATTRFLAG_*
+    char byteorder[8+1]; // String [012345rvwz] to define byte order
 
     entry()
       : raw_format(RAWFMT_DEFAULT),
         priority(PRIOR_DEFAULT),
         flags(0)
-      { }
+      { byteorder[0] = 0; }
   };
 
   entry & operator[](unsigned char id)
@@ -725,6 +743,10 @@ int ataReadSCTTempHist(ata_device * device, ata_sct_temperature_history_table *
 // Set SCT temperature logging interval
 int ataSetSCTTempInterval(ata_device * device, unsigned interval, bool persistent);
 
+// Get/Set SCT Error Recovery Control
+int ataGetSCTErrorRecoveryControltime(ata_device * device, unsigned type, unsigned short & time_limit);
+int ataSetSCTErrorRecoveryControltime(ata_device * device, unsigned type, unsigned short time_limit);
+
 
 /* Enable/Disable SMART on device */
 int ataEnableSmart (ata_device * device);
@@ -787,6 +809,9 @@ int isSupportSelectiveSelfTest(const ata_smart_values * data);
 inline bool isSCTCapable(const ata_identify_device *drive)
   { return !!(drive->words088_255[206-88] & 0x01); } // 0x01 = SCT support
 
+inline bool isSCTErrorRecoveryControlCapable(const ata_identify_device *drive)
+  { return ((drive->words088_255[206-88] & 0x09) == 0x09); } // 0x08 = SCT Error Recovery Control support
+
 inline bool isSCTFeatureControlCapable(const ata_identify_device *drive)
   { return ((drive->words088_255[206-88] & 0x11) == 0x11); } // 0x10 = SCT Feature Control support
 
@@ -803,7 +828,6 @@ enum ata_attr_state
 {
   ATTRSTATE_NON_EXISTING,   // No such Attribute
   ATTRSTATE_NO_NORMVAL,     // Normalized value not valid
-  ATTRSTATE_BAD_THRESHOLD,  // Threshold not valid
   ATTRSTATE_NO_THRESHOLD,   // Unknown or no threshold
   ATTRSTATE_OK,             // Never failed
   ATTRSTATE_FAILED_PAST,    // Failed in the past
@@ -812,8 +836,10 @@ enum ata_attr_state
 
 // Get attribute state
 ata_attr_state ata_get_attr_state(const ata_smart_attribute & attr,
-                                  const ata_smart_threshold_entry & thre,
-                                  const ata_vendor_attr_defs & defs);
+                                  int attridx,
+                                  const ata_smart_threshold_entry * thresholds,
+                                  const ata_vendor_attr_defs & defs,
+                                  unsigned char * threshval = 0);
 
 // Get attribute raw value.
 uint64_t ata_get_attr_raw_value(const ata_smart_attribute & attr,
