@@ -56,7 +56,7 @@
 #include "smartctl.h"
 #include "utility.h"
 
-const char * smartctl_cpp_cvsid = "$Id: smartctl.cpp 3046 2010-01-22 21:30:02Z chrfranke $"
+const char * smartctl_cpp_cvsid = "$Id: smartctl.cpp 3186 2010-10-16 13:09:11Z chrfranke $"
                                   CONFIG_H_CVSID EXTERN_H_CVSID SMARTCTL_H_CVSID;
 
 // This is a block containing all the "control variables".  We declare
@@ -90,6 +90,10 @@ void Usage (void){
 "         Show all SMART information for device\n\n"
 "  -x, --xall\n"
 "         Show all information for device\n\n"
+"  --scan\n"
+"         Scan for devices\n\n"
+"  --scan-open\n"
+"         Scan for devices and try to open each device\n\n"
   );
   printf(
 "================================== SMARTCTL RUN-TIME BEHAVIOR OPTIONS =====\n\n"
@@ -126,7 +130,7 @@ void Usage (void){
 "  -l TYPE, --log=TYPE\n"
 "        Show device log. TYPE: error, selftest, selective, directory[,g|s],\n"
 "                               background, sasphy[,reset], sataphy[,reset],\n"
-"                               scttemp[sts,hist],\n"
+"                               scttemp[sts,hist], scterc[,N,M],\n"
 "                               gplog,N[,RANGE], smartlog,N[,RANGE],\n"
 "                               xerror[,N][,error], xselftest[,N][,selftest]\n\n"
 "  -v N,OPTION , --vendorattribute=N,OPTION                            (ATA)\n"
@@ -138,15 +142,21 @@ void Usage (void){
 "        Drive-specific presets: use, ignore, show, showall\n\n"
 "  -B [+]FILE, --drivedb=[+]FILE                                       (ATA)\n"
 "        Read and replace [add] drive database from FILE\n"
-#ifdef SMARTMONTOOLS_DRIVEDBDIR
-"        [default is "SMARTMONTOOLS_DRIVEDBDIR"/drivedb.h]\n"
-#endif
-"\n"
+"        [default is +%s",
+    get_drivedb_path_add()
   );
+#ifdef SMARTMONTOOLS_DRIVEDBDIR
   printf(
+                      "\n"
+"         and then    %s",
+    get_drivedb_path_default()
+  );
+#endif
+  printf(
+         "]\n\n"
 "============================================ DEVICE SELF-TEST OPTIONS =====\n\n"
 "  -t TEST, --test=TEST\n"
-"        Run test. TEST: offline short long conveyance select,M-N\n"
+"        Run test. TEST: offline short long conveyance vendor,N select,M-N\n"
 "                        pending,N afterselect,[on|off] scttempint,N[,p]\n\n"
 "  -C, --captive\n"
 "        Do test in captive mode (along with -t)\n\n"
@@ -166,7 +176,7 @@ static std::string getvalidarglist(char opt)
   case 'q':
     return "errorsonly, silent, noserial";
   case 'd':
-    return smi()->get_valid_dev_types_str() + ", test";
+    return smi()->get_valid_dev_types_str() + ", auto, test";
   case 'T':
     return "normal, conservative, permissive, verypermissive";
   case 'b':
@@ -178,13 +188,14 @@ static std::string getvalidarglist(char opt)
   case 'S':
     return "on, off";
   case 'l':
-    return "error, selftest, selective, directory[,g|s], background, scttemp[sts|hist], "
+    return "error, selftest, selective, directory[,g|s], background, scttemp[sts|hist], scterc[,N,M], "
            "sasphy[,reset], sataphy[,reset], gplog,N[,RANGE], smartlog,N[,RANGE], "
 	   "xerror[,N][,error], xselftest[,N][,selftest]";
   case 'P':
     return "use, ignore, show, showall";
   case 't':
-    return "offline, short, long, conveyance, select,M-N, pending,N, afterselect,[on|off], scttempint,N[,p]";
+    return "offline, short, long, conveyance, vendor,N, select,M-N, "
+           "pending,N, afterselect,[on|off], scttempint,N[,p]";
   case 'F':
     return "none, samsung, samsung2, samsung3, swapid";
   case 'n':
@@ -221,6 +232,8 @@ enum checksum_err_mode_t {
 
 static checksum_err_mode_t checksum_err_mode = CHECKSUM_ERR_WARN;
 
+static void scan_devices(const char * type, bool with_open, const char * pattern);
+
 /*      Takes command options and sets features to be run */    
 const char * parse_options(int argc, char** argv,
                            ata_print_options & ataopts,
@@ -229,6 +242,7 @@ const char * parse_options(int argc, char** argv,
   // Please update getvalidarglist() if you edit shortopts
   const char *shortopts = "h?Vq:d:T:b:r:s:o:S:HcAl:iaxv:P:t:CXF:n:B:";
   // Please update getvalidarglist() if you edit longopts
+  enum { opt_scan = 1000, opt_scan_open = 1001 };
   struct option longopts[] = {
     { "help",            no_argument,       0, 'h' },
     { "usage",           no_argument,       0, 'h' },
@@ -258,6 +272,8 @@ const char * parse_options(int argc, char** argv,
     { "firmwarebug",     required_argument, 0, 'F' },
     { "nocheck",         required_argument, 0, 'n' },
     { "drivedb",         required_argument, 0, 'B' },
+    { "scan",            no_argument,       0, opt_scan      },
+    { "scan-open",       no_argument,       0, opt_scan_open },
     { 0,                 0,                 0, 0   }
   };
 
@@ -268,16 +284,14 @@ const char * parse_options(int argc, char** argv,
 
   const char * type = 0; // set to -d optarg
   bool no_defaultdb = false; // set true on '-B FILE'
+  int scan = 0; // set by --scan, --scan-open
   bool badarg = false, captive = false;
   int testcnt = 0; // number of self-tests requested
 
   int optchar;
   char *arg;
 
-  // This miserable construction is needed to get emacs to do proper indenting. Sorry!
-  while (-1 != (optchar = 
-                getopt_long(argc, argv, shortopts, longopts, NULL)
-                )){
+  while ((optchar = getopt_long(argc, argv, shortopts, longopts, 0)) != -1) {
     switch (optchar){
     case 'V':
       con->dont_print = false;
@@ -298,7 +312,7 @@ const char * parse_options(int argc, char** argv,
       }
       break;
     case 'd':
-      type = optarg;
+      type = (strcmp(optarg, "auto") ? optarg : (char *)0);
       break;
     case 'T':
       if (!strcmp(optarg,"normal")) {
@@ -430,6 +444,8 @@ const char * parse_options(int argc, char** argv,
         ataopts.sataphy = ataopts.sataphy_reset = true;
       } else if (!strcmp(optarg,"background")) {
         scsiopts.smart_background_log = true;
+      } else if (!strcmp(optarg,"scterc")) {
+        ataopts.sct_erc_get = true;
       } else if (!strcmp(optarg,"scttemp")) {
         ataopts.sct_temp_sts = ataopts.sct_temp_hist = true;
       } else if (!strcmp(optarg,"scttempsts")) {
@@ -467,6 +483,18 @@ const char * parse_options(int argc, char** argv,
         else
           badarg = true;
 
+      } else if (!strncmp(optarg, "scterc,", sizeof("scterc,")-1)) {
+        unsigned rt = ~0, wt = ~0; int n = -1;
+        sscanf(optarg,"scterc,%u,%u%n", &rt, &wt, &n);
+        if (n == (int)strlen(optarg) && rt <= 999 && wt <= 999) {
+          ataopts.sct_erc_set = true;
+          ataopts.sct_erc_readtime = rt;
+          ataopts.sct_erc_writetime = wt;
+        }
+        else {
+          sprintf(extraerror, "Option -l scterc,[READTIME,WRITETIME] syntax error\n");
+          badarg = true;
+        }
       } else if (   !strncmp(optarg, "gplog,"   , sizeof("gplog,"   )-1)
                  || !strncmp(optarg, "smartlog,", sizeof("smartlog,")-1)) {
         unsigned logaddr = ~0U; unsigned page = 0, nsectors = 1; char sign = 0;
@@ -526,6 +554,7 @@ const char * parse_options(int argc, char** argv,
       ataopts.smart_selective_selftest_log = true;
       ataopts.smart_logdir = ataopts.gp_logdir = true;
       ataopts.sct_temp_sts = ataopts.sct_temp_hist = true;
+      ataopts.sct_erc_get = true;
       ataopts.sataphy = true;
       scsiopts.smart_background_log = true;
       scsiopts.sasphy = true;
@@ -601,7 +630,8 @@ const char * parse_options(int argc, char** argv,
           ataopts.smart_selective_args.pending_time = i+1;
 	}
       } else if (!strncmp(optarg,"select",strlen("select"))) {
-        testcnt++;
+        if (ataopts.smart_selective_args.num_spans == 0)
+          testcnt++;
         // parse range of LBAs to test
         uint64_t start, stop; int mode;
         if (split_selective_arg(optarg, &start, &stop, &mode)) {
@@ -633,6 +663,15 @@ const char * parse_options(int argc, char** argv,
         }
         ataopts.sct_temp_int = interval;
         ataopts.sct_temp_int_pers = (n2 == len);
+      } else if (!strncmp(optarg, "vendor,", sizeof("vendor,")-1)) {
+        unsigned subcmd = ~0U; int n = -1;
+        if (!(   sscanf(optarg, "%*[a-z],0x%x%n", &subcmd, &n) == 1
+              && subcmd <= 0xff && n == (int)strlen(optarg))) {
+          strcpy(extraerror, "Option -t vendor,0xNN syntax error\n");
+          badarg = true;
+        }
+        else
+          ataopts.smart_selftest_type = subcmd;
       } else {
         badarg = true;
       }
@@ -675,6 +714,12 @@ const char * parse_options(int argc, char** argv,
       Usage();
       EXIT(0);  
       break;
+
+    case opt_scan:
+    case opt_scan_open:
+      scan = optchar;
+      break;
+
     case '?':
     default:
       con->dont_print = false;
@@ -726,6 +771,13 @@ const char * parse_options(int argc, char** argv,
       EXIT(FAILCMD);
     }
   }
+
+  // Special handling of --scan, --scanopen
+  if (scan) {
+    scan_devices(type, (scan == opt_scan_open), argv[optind]);
+    EXIT(0);
+  }
+
   // At this point we have processed all command-line options.  If the
   // print output is switchable, then start with the print output
   // turned off
@@ -828,21 +880,6 @@ void pout(const char *fmt, ...){
   return;
 }
 
-// This function is used by utility.cpp to report LOG_CRIT errors.
-// The smartctl version prints to stdout instead of syslog().
-void PrintOut(int priority, const char *fmt, ...) {
-  va_list ap;
-
-  // avoid warning message about unused variable from gcc -W: just
-  // change value of local copy.
-  priority=0;
-
-  va_start(ap,fmt);
-  vprintf(fmt,ap);
-  va_end(ap);
-  return;
-}
-
 // Used to warn users about invalid checksums. Called from atacmds.cpp.
 // Action to be taken may be altered by the user.
 void checksumwarning(const char * string)
@@ -869,9 +906,50 @@ static const char * get_protocol_info(const smart_device * dev)
   }
 }
 
+// Device scan
+// smartctl [-d type] --scan[-open] [PATTERN]
+void scan_devices(const char * type, bool with_open, const char * pattern)
+{
+  bool dont_print = !(con->reportataioctl || con->reportscsiioctl);
+  smart_device_list devlist;
+
+  con->dont_print = dont_print;
+  bool ok = smi()->scan_smart_devices(devlist, type , pattern);
+  con->dont_print = false;
+
+  if (!ok) {
+    pout("scan_smart_devices: %s\n", smi()->get_errmsg());
+    return;
+  }
+
+  for (unsigned i = 0; i < devlist.size(); i++) {
+    smart_device * dev = devlist.at(i);
+
+    std::string openmsg;
+    if (with_open) {
+      con->dont_print = dont_print;
+      dev = dev->autodetect_open();
+      con->dont_print = false;
+
+      if (dev->is_open())
+        openmsg = " (opened)";
+      else
+        openmsg = strprintf(" (open failed: %s)", dev->get_errmsg());
+    }
+
+    pout("%s -d %s [%s]%s\n", dev->get_info_name(), dev->get_dev_type(),
+         get_protocol_info(dev), openmsg.c_str());
+    if (dev->is_open())
+      dev->close();
+  }
+}
+
 // Main program without exception handling
 int main_worker(int argc, char **argv)
 {
+  // Throw if CPU endianess does not match compile time test.
+  check_endianness();
+
   // Initialize interface
   smart_interface::init();
   if (!smi())
